# -*- coding: utf-8 -*-

"""Classes for validating data supplied to the application."""
from __future__ import unicode_literals

from hyputils.memex.schemas.base import ValidationError

__all__ = ("ValidationError",)
